#!/usr/bin/wperl
use strict;
use warnings;
#wikiperl.pl
# 	Applies an arbitrary amount of perl code to process a list of pages.
# 	Code is released under the terms of the GPL.

#Configuration options
#	Change these to modify default behavior
our %_GLOBALOPTIONS = (
		#set the location of the login config file. Uses the current directory
		#by default if not specified
	CONFIGFILE => "",
		#1 = open target window on program startup
	OPENTARGETWINDOW => 1,
		#1 = open init code window on program startup
	OPENINITWINDOW => 0,
		#1 = open console window on program startup
	OPENCONSOLEWINDOW => 1,
		#Colors for the various edit windows. This beats a basic black-on-white scheme.
	EDITTEXTBACKGROUND => "#000000",	#editable text background
	EDITTEXTFOREGROUND => "#AAAAAA",	#editable text font color
	LOCKTEXTBACKGROUND => "#AAAAAA",	#text that is locked during transform, background
	LOCKTEXTFOREGROUND => "#000000",	#locked text, font color
	ROTEXTBACKGROUND   => "#000033",	#read-only text background color
	ROTEXTFOREGROUND   => "#999999",	#read-only text font color
	ERRORPOPUP         => 1,
	ERRORCONSOLE       => 1, 
);

#DO NOT MODIFY BELOW THIS LINE
#	Options above this line may be modified by the user.
#	Options below this line will effect the operation of the program.

#Libraries
our $VERSION = "Alpha 3";
use Tk;
use Tk::DialogBox;
use Tk::ROText;
use Tk::ProgressBarPlus;
use MediaWiki;
use XML::Simple;

#Global Variables
our $_init = "";
our %_opts = (
	"input"    => 0,
	"output"   => 2,
	"halt"     => 0,
	"delay"    => 0,
	"maxtest"  => 5,
	"loggedin" => 0, 
	"username" => "",
	"server"   => "",
);
our @_handles = ();

#Start the program.
my %targetwin = ("window" => undef, "pages" => undef, "basepage" => undef);
my %mainwin   = ("window" => undef, "transform" => undef, "summary" => undef, "progbar" => undef);
my %conwin    = ("window" => undef, "console" => undef, "main" => \%mainwin);
my %initwin   = ("window" => undef);
my $filename;
my $mw = MediaWiki->new();
$mw->setup(
	{
		'bot'  => {},
		'wiki' => {
				'host' => "",
				'path' => 'w'
		}
	}
);
make_main_window(\%mainwin, \%targetwin, \%conwin, \%initwin, $mw, \$filename);
make_init_window(\%mainwin, \%initwin, \%conwin) if $_GLOBALOPTIONS{"OPENINITWINDOW"} == 1;
make_target_window(\%mainwin, \%targetwin, \%conwin, $mw) if $_GLOBALOPTIONS{"OPENTARGETWINDOW"} == 1; 
make_console_window(\%mainwin, \%conwin, $mw) if $_GLOBALOPTIONS{"OPENCONSOLEWINDOW"} == 1; 
print_console_message(\%conwin, "Program started");
MainLoop();

sub make_main_window
{
	#Create first grid, for edit window and scroll bars.
	our %_GLOBALOPTIONS;
	my ($mainwin, $targetwin, $conwin, $initwin, $mw, $filename) = @_;
	$mainwin->{'window'} = MainWindow->new();
	my $textframe = $mainwin->{'window'}->Frame->grid(
		-row    => 1, 
		-column => 1, 
		-sticky => "nsew"
	);
	$mainwin->{"transform"} = $textframe->Text(
		-wrap => "none",
		-background       => $_GLOBALOPTIONS{"EDITTEXTBACKGROUND"}, 
		-foreground       => $_GLOBALOPTIONS{"EDITTEXTFOREGROUND"},
		-insertbackground => $_GLOBALOPTIONS{"EDITTEXTFOREGROUND"}
	)->grid(
		-row    => 1,
		-column => 1,
		-sticky => "nsew"
	);
	my $srl_y = $textframe->Scrollbar(	
		-orient  => 'v',
		-command => [yview => $mainwin->{"transform"}]
	)->grid(
		-row    => 1,
		-column => 2,
		-sticky => "ns"
	);
	my $srl_x = $textframe->Scrollbar(
		-orient  => 'h',
		-command => [xview => $mainwin->{"transform"}]
	)->grid(
		-row    => 2,
		-column => 1,
		-sticky => "ew"
	);
	$mainwin->{"transform"}->configure(
		-yscrollcommand => ['set', $srl_y], 
		-xscrollcommand => ['set', $srl_x]
	);
	$mainwin->{"transform"}->bind('<Control-Key-s>' => \sub 
		{
			save_file($filename, $mainwin, $targetwin, $conwin);
			$mainwin{"transform"}->break();
		}
	);	
	my (@tags) = $mainwin->{"transform"}->bindtags();
	$mainwin->{"transform"}->bindtags([@tags[1, 0]]);
	$mainwin->{"summary"} = $textframe->Text(
		-height => 1, 
		-wrap   => "none",
		-background       => $_GLOBALOPTIONS{"EDITTEXTBACKGROUND"}, 
		-foreground       => $_GLOBALOPTIONS{"EDITTEXTFOREGROUND"},
		-insertbackground => $_GLOBALOPTIONS{"EDITTEXTFOREGROUND"}
	)->grid(
		-row        => 3, 
		-column     => 1, 
		-columnspan => 2, 
		-sticky     => "ew"
	);
	$mainwin->{"summary"}->bind('<Control-Key-s>' => \sub 
		{
			save_file($filename, $mainwin, $targetwin, $conwin);
			$mainwin{"summary"}->break();
		}
	);
	(@tags) = $mainwin->{"summary"}->bindtags();
	$mainwin->{"summary"}->bindtags([@tags[1, 0]]);

	#Create lower frame for buttons and progress bar
	my $lower = $mainwin->{'window'}->Frame->grid(
		-row    => 2, 
		-column => 1, 
		-sticky => "ew"
	);
	my $bframe = $lower->Frame->grid( 
		-column => 1, 
		-row    => 1, 
		-sticky => "w"
	);
	$bframe->Button(
		-text    => "Post",   
		-command => [\&post_transform, $mainwin, $targetwin, $conwin, $mw]
	)->pack(-side => "left");
	$bframe->Button(
		-text    => "Clear",  
		-command => [\&clear_program, $mainwin, $conwin]
	)->pack(-side => "left");
	$bframe->Button(
		-text    => "Stop",   
		-command => \sub{$_opts{'halt'} = 1;}
	)->pack(-side => "left");
	$mainwin->{'progbar'} = $lower->ProgressBarPlus(
		-blocks      => 100, 
		-gap         => 0, 
		-colors      => [0, "grey55", 100], 
		-troughcolor => "blue", 
		-showvalue   => 1, 
		-valueformat => '%.5s %%'
	)->grid(
		-column => 2, 
		-row    => 1, 
		-sticky => "nsew", 
		-padx   => 10
	);

	#Configure frames
	$mainwin->{'progbar'}->value(100);
	$textframe->gridColumnconfigure(1, -weight => "1");
	$textframe->gridRowconfigure(1,    -weight => "1");
	$lower->gridColumnconfigure(2, 	   -weight => "1");
	$mainwin->{'window'}->gridColumnconfigure(1, -weight => "1");
	$mainwin->{'window'}->gridRowconfigure(1,    -weight => "1");
	
	#Create Menu
	my $top = $mainwin->{'window'}->toplevel;
	my $menu = $top->Menu(-type => "menubar");
	$top->configure(-menu => $menu);
	$menu->Cascade(-label => "~File", -tearoff => 0, -menuitems => [
		[Button => 'Post',    -command =>[\&post_transform, $mainwin, $targetwin, $conwin, $mw]],
		[Separator => ''],
		[Button => 'Open',    -command => [\&load_file, $mainwin, $targetwin, $conwin]],
		[Button => 'Save',    -command => [\&save_file, $filename, $mainwin, $targetwin, $conwin]],
		[Button => 'Save as', -command => [\&save_as_file, $filename, $mainwin, $targetwin, $conwin]],
		[Separator => ''],
		[Button => 'Clear',   -command => [\&clear_program, $mainwin, $conwin]],
		[Button => 'Exit',    -command => [\&quit_program, $mw]]
	]);
	$menu->Cascade(-label => "~View", -tearoff => 0, -menuitems => [
		[Button => 'Init Code',    -command => [\&make_init_window, $mainwin, $initwin, $conwin]],
		[Button => 'Page Targets', -command => [\&make_target_window, $mainwin, $targetwin, $conwin, $mw]],
		[Button => 'Console',      -command => [\&make_console_window, $mainwin, $conwin]]
	]);
	$menu->Cascade(-label => "~Options", -tearoff => 0, -menuitems => [
		[Button => 'Options', -command => [\&make_options_window, $mainwin, $conwin]],
		[Button => 'Login',   -command => [\&make_login_window, $mainwin, $conwin, $mw]],
		[Button => 'Logout',  -command => [\&logout_from_server, $mainwin, $conwin, $mw]]
	]);
	$menu->Cascade(-label => "~Mode", -tearoff => 1, -menuitems => [
		[Separator => ''],
		[Radiobutton => "From Server", -value => 0, -variable => \${_opts{'input'}}],
		[Radiobutton => "From File",   -value => 1, -variable => \${_opts{'input'}}],
		[Radiobutton => "No Input",    -value => 2, -variable => \${_opts{'input'}}],
		[Separator => ''],
		[Radiobutton => "To Server", -value => 0, -variable => \${_opts{'output'}}],
		[Radiobutton => "To File",   -value => 1, -variable => \${_opts{'output'}}],
		[Radiobutton => "Test Edit", -value => 2, -variable => \${_opts{'output'}}],
		[Radiobutton => "No Output", -value => 3, -variable => \${_opts{'output'}}],
	]);
	$menu->Cascade(-label => "~Help", -tearoff => 0, -menuitems => [
		[Button => "About", -command => [\&make_about_window, $mainwin]]
	]);
}

#Apply current transformation to the current pagelist
sub post_transform
{
	our %_opts;
	our $_init;
	our %_GLOBALOPTIONS;
	my ($mainwin, $targetwin, $conwin, $mw) = @_;
	my $filename;
	if($_opts{'loggedin'} == 0 && ($_opts{'input'} == 0 || $_opts{'output'} == 0))
	{
		print_console_message($conwin, 'Error: Not connected to a server');
		$mainwin->{'window'}->messageBox(
			-message => 'Not connected to a server', 
			-type => 'ok', 
			-icon => 'error'
		);
		return;
	}
	print_console_message($conwin, "Initializing Transform...");
	my $starttime = time();			#get the time that we start
	my $transform = $mainwin->{'transform'}->Contents;
	chomp $transform;
	my $summary = $mainwin->{'summary'}->Contents;
	chomp $summary;
	my @pages = get_target_pages($targetwin);
	eval($_init) unless ($_init eq "");
	return unless alert_error($conwin, $@, "Init Error: " . $@);
	my $func = eval('return sub{$_=shift;my($page,$next,$last)=@_;' . $transform . ';return $_;}');
	return unless alert_error($conwin, $@, "Eval Error: " . $@);
	if(scalar @pages == 0 && $_opts{'input'} != 2)		#if no pages listed, error and return
	{
		alert_error($conwin, "1", "No pages listed");
		return;
	}
	elsif($_opts{'input'} == 2)
	{
		@pages = ('(null)');
	}
	$mainwin->{'progbar'}->value(0);
	lock_text_widgets($mainwin->{'transform'}, $mainwin->{'summary'});
	$mainwin->{'window'}->update();
	for(my $i = 0; $i < scalar @pages; $i++)
	{							#Loop through all pages in page list.
		if($_opts{'halt'} == 1)
		{
			print_console_message($conwin, "Halted");
			$mainwin->{'progbar'}->value(100);
			unlock_text_widgets($mainwin->{'transform'}, $mainwin->{'summary'});
			$_opts{'halt'} = 0;
			return;
		}
		my ($page, $pg, $next, $last, $text) = ($pages[$i], undef, "", "", "");
		$next = $pages[$i + 1] unless $i == scalar @pages - 1;
		$last = $pages[$i - 1] unless $i == 0;
		print_console_message($conwin, "Transforming '$page'");
		if($_opts{'input'} == 0) #input from wiki
		{
			$pg = $mw->get($pages[$i]);
			$text = $pg->content();
		}
		elsif($_opts{'input'} == 1) #input from file
		{
			$filename = $page;
			$filename =~ tr/:\//-_/;
			if(-e $filename . ".wiki")
			{
				open INPUT, "<$filename.wiki";
				$text = join('', <INPUT>);
				close INPUT;
			}
		}
		elsif($_opts{'input'} == 2) #no input
		{
			$text = '';
		}
		if($_opts{'output'} == 0) #output to wiki
		{
			$pg = $mw->get($page) unless defined $pg;
			$pg->{'content'} = $func->($text, $page, $next, $last);
			$pg->{'summary'} = $summary;
			$pg->save();
		}
		elsif($_opts{'output'} == 1) #output to file
		{
			$filename = $page;
			$filename =~ tr/:\//-_/;
			$text = $func->($text, $page, $next, $last);
			open OUTPUT, ">$filename.wiki";
			print OUTPUT $text;
			close OUTPUT;
		}
		elsif($_opts{'output'} == 2) #test output
		{
			last if($i >= $_opts{'maxtest'});
			$text = $func->($text, $page, $next, $last);
			show_test($mainwin, $page, $text);
		}
		elsif($_opts{'output'} == 3) #no output, run function for sideeffects
		{
			$func->($text, $page, $next, $last);
		}
		sleep $_opts{'delay'} if $_opts{'delay'};	
		$mainwin->{'progbar'}->value((($i + 1) / (scalar @pages)) * 100);
		print_console_message($conwin, "Finished " . ($i + 1) . " / " . scalar @pages );
	}
	unlock_text_widgets($mainwin->{'transform'}, $mainwin->{'summary'});
	$mainwin->{'progbar'}->value(100);
	my $endtime = time();
	print_console_message($conwin, "Done in " . ($endtime - $starttime) . " seconds");
	print_console_message($conwin, "Average " . 
		sprintf("%2.3f", (($endtime - $starttime) / (scalar @pages))) .
		" seconds per page"); 
}

sub alert_error
{
	my ($conwin, $flag, $msg) = @_;
	return 1 unless $flag;
	print_console_message($conwin, $msg) if($_GLOBALOPTIONS{'ERRORCONSOLE'} == 1);
	$conwin->{'main'}->{'window'}->messageBox(
		-message => $msg, 
		-type => 'ok', 
		-icon => 'error'
	) if($_GLOBALOPTIONS{'ERRORPOPUP'} == 1);
	return 0;
}

sub unlock_text_widgets
{
	our %_GLOBALOPTIONS;
	foreach my $w (@_)
	{
		$w->configure(
			-state => "normal", 
			-background => $_GLOBALOPTIONS{"EDITTEXTBACKGROUND"},
			-foreground => $_GLOBALOPTIONS{"EDITTEXTFOREGROUND"}
		);
	}
}

sub lock_text_widgets
{
	our %_GLOBALOPTIONS;
	foreach my $w (@_)
	{
		$w->configure(
			-state => "disabled", 
			-background => $_GLOBALOPTIONS{"LOCKTEXTBACKGROUND"},
			-foreground => $_GLOBALOPTIONS{"LOCKTEXTFOREGROUND"}
		);
	}
}

#Clear the program windows
sub clear_program
{
	our ($_init, %_opts, @_handles);
	my ($mainwin, $conwin) = shift;
	$_init = "";
	$_opts{'input'} = 0;
	$_opts{'output'} = 2;
	$mainwin->{'transform'}->Contents("");
	$mainwin->{'summary'}->Contents("");
	foreach my $handle (@_handles)
	{
		$handle->destroy() if Exists($handle);
	}
	@_handles = ();
	clear_console_window($conwin);
}

#exit the program, may include options to "save current task?"
sub quit_program
{
	my $mw = shift;
	$mw->logout();
	exit(0);
}

#load a task from a file.
sub load_file
{
	our $_init;
	my ($mainwin, $targetwin, $conwin) = @_;
	my $types = [
		['XML Files',			'.xml'],
		['Filter Files',        '.xml']
	];
	my $filename = $mainwin->{'window'}->getOpenFile(-filetypes => $types);
	if($filename ne "")
	{
		chomp $filename;
		print "opening " . $filename . "\n";
		my $in = XMLin($filename, 'ForceArray' => 1);
		$mainwin->{'summary'}->Contents($in->{'summary'}->[0]);
		$_init = $in->{'code'}->[0]->{'init'}->[0];
		my $c = $in->{'code'}->[0]->{'transform'}->[0];
		$c =~ s/__LT__/</g;
		$c =~ s/__GT__/>/g;
		$c =~ s/__AMP__/&/g;
		$mainwin->{'transform'}->Contents($c);
		if(Exists($targetwin->{'window'}))
		{
			if(defined($in->{'pages'}->[0]->{'page'}))
			{
				my $p = @{$in->{'pages'}->[0]->{'page'}};
				$targetwin->{'pages'}->Contents(join("\n", $p));
			}
			if(defined($in->{'pages'}->[0]->{'basepage'}->[0]))
			{
				$targetwin->{'basepage'}->Contents($in->{'pages'}->[0]->{'basepage'}->[0]);
			}
		}
		print_console_message($conwin, "Opened $filename");
	}
}

#Allow the user to save a file, selecting a filename. 
sub save_as_file
{
	my ($filename, $mainwin, $targetwin, $conwin) = @_;
	$$filename = $mainwin->{'window'}->getSaveFile(-defaultextension => ".xml");
	save_file($filename, $mainwin, $targetwin, $conwin) if defined $filename && $filename ne "";
}

#save the current task to a file.
sub save_file
{
	our $_init;
	my ($filename, $mainwin, $targetwin, $conwin) = @_;
	
	if(defined($$filename) && $$filename ne "")
	{
		chomp $$filename;
		my $output = << "_END_XML_";
<?xml version=\"1.0\" encoding=\"UTF-8\"?>
<!DOCTYPE filter [
	<!ELEMENT filter (summary?, pages, code)>
	<!ELEMENT pages (page*, basepage?)>
	<!ELEMENT code (init?, transform?)>
	<!ELEMENT init (#PCDATA)>
	<!ELEMENT summary (#PCDATA)>
	<!ELEMENT page (#PCDATA)>
	<!ELEMENT basepage (#PCDATA)>
	<!ELEMENT transform (#PCDATA)>
]>

_END_XML_

		$output .= "<filter version=\"1.0\">\n<code>\n";
		$output .= "<init>" . $_init . "</init>\n" unless (defined($_init) && $_init eq "");
		my $c = $mainwin->{'transform'}->Contents;
		chomp $c;
		$c =~ s/</__LT__/g;
		$c =~ s/>/__GT__/g;
		$c =~ s/&/__AMP__/g;
		$output .= "<transform>" . $c. "</transform>\n" unless $c eq "";
		$output .= "</code>\n";
		my $s = $mainwin->{'summary'}->Contents;
		chomp $s;
		$output .= "<summary>" . $s . "</summary>\n" unless $s eq "";
		if(Exists($targetwin->{'window'}))
		{
			$output .= "<pages>\n";
			my $b = $targetwin->{'basepage'}->Contents;
			chomp $b;
			$output .= "<basepage>" . $b . "</basepage>\n" unless $b eq '';
			foreach my $p (get_target_pages($targetwin))
			{
				$output .= "<page>" . $p . "</page>\n";
			}
			$output .= "</pages>\n";
		}
		$output .= "</filter>";
		open OUTPUT, ">$$filename";
		print OUTPUT $output;
		close(OUTPUT);
		print_console_message($conwin, "Saved file '$$filename'");
	}
	else
	{
		save_as_file(@_);
		return;
	}
}

#Display an initialization code window
sub make_init_window
{
	our $_init;
	our %_GLOBALOPTIONS;
	my($mainwin, $initwin, $conwin) = @_;
	my $statmsg;	#status message string
	return if Exists($initwin->{'window'});
	$initwin->{'window'} = $mainwin->{'window'}->Toplevel;
	$initwin->{'window'}->title("Initialization Code");
	my $textframe = $initwin->{'window'}->Frame;
	$initwin->{'init'} = $textframe->Text(
		-width  => 60, 
		-height => 15, 
		-wrap   => "none",
		-background       => $_GLOBALOPTIONS{"EDITTEXTBACKGROUND"}, 
		-foreground       => $_GLOBALOPTIONS{"EDITTEXTFOREGROUND"},
		-insertbackground => $_GLOBALOPTIONS{"EDITTEXTFOREGROUND"}
	)->grid(
		-row    => 1, 
		-column => 1, 
		-sticky => "nsew"
	);
	$initwin->{'init'}->bind('<KeyPress>' => sub { $statmsg = ""; });
	my $srl_y = $textframe->Scrollbar(
		-orient  => 'v',
		-command => [yview => $initwin->{'init'}]
	)->grid(
		-row    => 1, 
		-column => 2, 
		-sticky => "ns"
	);
	my $srl_x = $textframe->Scrollbar(
		-orient  => 'h',
		-command => [xview => $initwin->{'init'}]
	)->grid(
		-row    => 2, 
		-column => 1,
		-sticky => "ew"
	);
	$initwin->{'init'}->configure(
		-yscrollcommand => ['set', $srl_y], 
		-xscrollcommand => ['set', $srl_x]
	);
	$initwin->{'init'}->Contents($_init);
	my $frame = $initwin->{'window'}->Frame;
	$frame->Button(
		-text    => "Save", 
		-command => \sub
		{
			$_init = $initwin->{'init'}->Contents(); 
			chomp $_init;
			$statmsg = "Init code saved";
			print_console_message($conwin, "Init code updated");
		}
	)->pack(-side => "left");
	$frame->Button(
		-text    => "Close", 
		-command => \sub
		{
			$initwin->{'window'}->destroy();
			$statmsg = "";
		}
	)->pack(-side => "left");
	$frame->Label(-textvariable => \$statmsg)->pack(-side => "left");
	$textframe->grid(
		-row    => 1, 
		-column => 1, 
		-sticky => "nsew"
	);
	$textframe->gridRowconfigure(1,    -weight=>1);
	$textframe->gridColumnconfigure(1, -weight=>1);
	$frame->grid(-row=>2, -column=>1,  -sticky=>"sw");
	$initwin->{'window'}->gridColumnconfigure(1, -weight=>1);
	$initwin->{'window'}->gridRowconfigure(1,    -weight=>1);
}

#Displays a message window to display messages
sub make_console_window
{
	our %_GLOBALOPTIONS;
	my ($mainwin, $conwin) = @_;
	return if Exists($conwin->{'window'});
	$conwin->{'window'} = $mainwin->{'window'}->Toplevel;
	$conwin->{'window'}->title("Message Console");
	my $textframe = $conwin->{'window'}->Frame;
	$conwin->{'console'} = $textframe->ROText(
		-width  => 60, 
		-height => 15, 
		-wrap   => "none",
		-background => $_GLOBALOPTIONS{"ROTEXTBACKGROUND"}, 
		-foreground => $_GLOBALOPTIONS{"ROTEXTFOREGROUND"}
	)->grid(
		-row    => 1, 
		-column => 1, 
		-sticky => "nsew"
	);
	my $srl_y = $textframe->Scrollbar(
		-orient  => 'v',
		-command => [yview => $conwin->{'console'}]
	)->grid(
		-row    => 1, 
		-column => 2, 
		-sticky => "ns"
	);
	my $srl_x = $textframe->Scrollbar(
		-orient  => 'h',
		-command => [xview => $conwin->{'console'}]
	)->grid(
		-row    => 2, 
		-column => 1,
		-sticky => "ew"
	);
	$conwin->{'console'}->configure(
		-yscrollcommand => ['set', $srl_y], 
		-xscrollcommand => ['set', $srl_x]
	);
	my $frame = $conwin->{'window'}->Frame;
	
	$frame->Button(
		-text    => "Clear", 
		-command => [\&clear_console_window, $conwin]
	)->pack(-side => "left");
	$frame->Button(
		-text    => "Close", 
		-command => \sub { $conwin->{'window'}->destroy(); }
	)->pack(-side => "left");
	$textframe->grid(
		-row    => 1, 
		-column => 1, 
		-sticky => "nsew"
	);
	$textframe->gridRowconfigure(1,    -weight => 1);
	$textframe->gridColumnconfigure(1, -weight => 1);
	$frame->grid(-row => 2, -column => 1, -sticky => "sw");
	$conwin->{'window'}->gridColumnconfigure(1, -weight => 1);
	$conwin->{'window'}->gridRowconfigure(1,    -weight => 1);
}

#prints a message to the console window, if it's open.
sub print_console_message
{
	my ($conwin, $message) = @_;
	if(Exists($conwin->{'window'}))
	{
		#I want to do some kind of buffering or truncating of messages here
		#so that the console doesnt just keep growning.
		my $existingtext = $conwin->{'console'}->Contents;
		chomp $existingtext;
		$conwin->{'console'}->Contents("> " . $message . "\n" . $existingtext);
	}
	$conwin->{'main'}->{'window'}->update();
}

#clears the console window
sub clear_console_window
{
	my $conwin = shift;
	return unless Exists($conwin->{'window'});
	$conwin->{'console'}->Contents("> Cleared");
}

#Displays a window to set program options.
sub make_options_window
{
	our %_opts;
	my ($mainwin, $conwin) = @_;
	my $options = $mainwin->{'window'}->Toplevel;
	my $frame = $options->Frame;
	$frame->Entry(
		-width        => 5, 
		-textvariable => \$_opts{'delay'}
	)->pack(-side => "left");
	$frame->Label(-text => "Delay (sec)")->pack();
	$frame->pack(-anchor => "w");
	my $frame2 = $options->Frame;
	$frame2->Entry(
		-width        => 5, 
		-textvariable => \$_opts{'maxtest'}
	)->pack(-side => "left");
	$frame2->Label(-text  => "Max test windows")->pack();
	$frame2->pack(-anchor => "w");
	$options->Label(-text => "Username: " . $_opts{"username"})->pack();
	$options->Label(-text => "Server: " . $_opts{"server"})->pack();
	$options->Button(
		-text    => "Close", 
		-command => [sub
			{
				print_console_message($conwin, "Delay: " . $_opts{"delay"});
				print_console_message($conwin, "Max test windows: " . $_opts{"maxtest"});
				print_console_message($conwin, "Updated options");
				$options->destroy();
			}
		]
	)->pack(-anchor => "w");
	$options->title('Options');
}

#displays a test window with no return handle. That is, it cannot be modified once created.
sub show_test
{
	our %_GLOBALOPTIONS;
	our @_handles;
	my ($mainwin, $title, $text) = @_;
	my $testwin = $mainwin->{'window'}->Toplevel;
	my $frame = $testwin->Frame->grid(
		-row    => 1,
		-column => 1,
		-sticky => "nsew"
	);
	my $ttest = $frame->ROText(
		-width  => 80, 
		-height => 20,
		-background => $_GLOBALOPTIONS{"ROTEXTBACKGROUND"}, 
		-foreground => $_GLOBALOPTIONS{"ROTEXTFOREGROUND"}
	)->grid(
		-row    => 1,
		-column => 1,
		-sticky => "nsew"
	);
	my $srl_y = $frame->Scrollbar(
		-orient  => 'v',
		-command => [yview => $ttest],
	)->grid(
		-row    => 1,
		-column => 2,
		-sticky => "ns"
	);
	$ttest->configure(-yscrollcommand => ['set', $srl_y]);
	$ttest->Contents($text);
	$testwin->Button(
		-text    => "Close",  
		-command => [$testwin => 'destroy']
	)->grid(
		-row    => 3,
		-column => 1,
		-sticky => "w"
	); 
	$testwin->title("Test: " . $title);
	$frame->gridRowconfigure(1,      -weight => 1);
	$frame->gridColumnconfigure(1,   -weight => 1);
	$testwin->gridColumnconfigure(1, -weight => 1);
	$testwin->gridRowconfigure(1,    -weight => 1);
	push @_handles, $testwin;
}
	
sub make_login_window
{	
	our %_opts;
	my ($mainwin, $conwin, $mw) = @_;
	my $box = $mainwin->{'window'}->DialogBox(
		-title   => "Login", 
		-buttons => ["Login", "Anon", "Cancel"]
	);
	$box->add("Label", -text => "Username:")->pack();
	my $name = $box->add("Text", 
		-width  => 30, 
		-height => 1, 
		-wrap   => "none"
	)->pack();		
	$box->add("Label", -text => "Password:")->pack();
	my $pass = $box->add("Text", 
		-width      => 30, 
		-height     => 1, 
		-wrap       => "none", 
		-foreground => "white"
	)->pack();
	$box->add("Label", -text => "Server:")->pack();
	my $serv = $box->add("Text", 
		-width  => 30, 
		-height => 1, 
		-wrap   => "none"
	)->pack();
	my $hijacktab = sub 
	{
		my ($this, $next) = @_;
		$this->bind('<Key-Tab>' => \sub
		{
			$next->focus();
			$this->break();
		});
		my @tags = $this->bindtags();
		$this->bindtags([@tags[1, 0]]);
	};
	$hijacktab->($name, $pass);
	$hijacktab->($pass, $serv);
	$hijacktab->($serv, $name);
	my ($username, $server) = get_login_config();
	$name->Contents($username);
	$serv->Contents($server);
	my $d = $box->Show;
	$username = $name->Contents;
	chomp $username;
	my $password = $pass->Contents;
	chomp $password;
	$server = $serv->Contents;
	chomp $server;
	$server = undef unless $server;
	my $dologin = sub
	{
		my ($u, $p, $s) = @_;
		my $stat = 1;
		$stat = 0 unless defined($s);
		$mw->switch($s);
		$mw->login($u, $p) if defined($p);
		$mainwin->{'window'}->title("Wikiperl - $u - $s");
		print_console_message($conwin, "Editing " . (defined($s)?($s):("No Server")) . " as $u");
		$_opts{'loggedin'} = $stat;
		$_opts{'server'}   = $s;
		$_opts{'username'} = $u;
		save_login_config($u, $s) if defined($p);
	};
	if($d eq "Login")
	{
		$dologin->($username, $password, $server);
	}
	elsif($d eq "Anon")
	{
		$dologin->("Anonymous", undef, $server);
	}
	#come here if clicked "cancel". no change from status quo
	return;
}

sub get_login_config
{
	our $_CONFIGFILE;
	my $configfile;
	if(!defined($_CONFIGFILE) || $_CONFIGFILE eq "")
	{
		$configfile = "wp_config.xml";
	}
	else
	{
		$configfile = $_CONFIGFILE;
	}
	return ("", "") unless -e $configfile;
	my $in = XMLin($configfile, 'ForceArray' => 1);
	my $username = $in->{'username'}->[0];
	my $server = $in->{'server'}->[0];
	return ($username, $server);
}

sub save_login_config
{
	our $_CONFIGFILE;
	my $configfile;
	if(!defined($_CONFIGFILE) || $_CONFIGFILE eq "")
	{
		$configfile = "wp_config.xml";
	}
	else
	{
		$configfile = $_CONFIGFILE;
	}
	my ($username, $server) = @_;
	my $out = << "__END_OF_XML__";
<?xml version=\"1.0\" encoding=\"UTF-8\"?>
<!DOCTYPE wploginconfig [
	<!ELEMENT wploginconfig (server, username)>
	<!ELEMENT server (#PCDATA)>
	<!ELEMENT username (#PCDATA)>
]>
<wploginconfig>
	<server>$server</server>
	<username>$username</username>
</wploginconfig>

__END_OF_XML__

	open OUTPUT, ">$configfile";
	print OUTPUT "$out";
	close OUTPUT;
}

sub logout_from_server
{
	our %_opts;
	my ($mainwin, $conwin, $mw) = @_;
	$mw->logout();
	$mainwin->{'window'}->title('Wikiperl - Anonymous - No Server');
	print_console_message($conwin, "Logged out. Not connected to a server");
	$_opts{"loggedin"} = 0;
}

#Generate a new targets window, if one does not already exist
sub make_target_window
{
	our %_GLOBALOPTIONS;
	my ($mainwin, $targetwin, $conwin, $mw) = @_;
	return if defined($targetwin->{'window'});
	$targetwin->{'window'} = $mainwin->{'window'}->Toplevel;

	my $textframe = $targetwin->{'window'}->Frame();
	$targetwin->{'pages'} = $textframe->Text(
		-width  => 80, 
		-height => 15, 
		-wrap   => "none",
		-background       => $_GLOBALOPTIONS{"EDITTEXTBACKGROUND"}, 
		-foreground       => $_GLOBALOPTIONS{"EDITTEXTFOREGROUND"},
		-insertbackground => $_GLOBALOPTIONS{"EDITTEXTFOREGROUND"}
	)->grid(
		-row    => 2,
		-column => 1, 
		-sticky => "nsew"
	);
	my $srl_y = $textframe->Scrollbar(
		-orient  => 'v',
		-command => [yview => $targetwin->{'pages'}]
	)->grid(
		-row    => 2, 
		-column => 2,
		-sticky => "ns"
	);
    my $srl_x = $textframe->Scrollbar(
		-orient  => 'h',
		-command => [xview => $targetwin->{'pages'}]
	)->grid(
		-row    => 3, 
		-column => 1, 
		-sticky => "ew"
	);
    $targetwin->{'pages'}->configure(
		-yscrollcommand => ['set', $srl_y], 
		-xscrollcommand => ['set', $srl_x]
	);
	$textframe->Label(-text => "List page:")->grid(
		-row    => 4, 
		-column => 1, 
		-sticky => "sw"
	);
	$targetwin->{'basepage'} = $textframe->Text(
		-width  => 80, 
		-height => 1, 
		-wrap   => "none",
		-background       => $_GLOBALOPTIONS{"EDITTEXTBACKGROUND"}, 
		-foreground       => $_GLOBALOPTIONS{"EDITTEXTFOREGROUND"},
		-insertbackground => $_GLOBALOPTIONS{"EDITTEXTFOREGROUND"}
	)->grid(
		-row        => 5, 
		-column     => 1, 
		-columnspan => 2, 
		-sticky     => "sew"
	);
	$textframe->gridColumnconfigure(1, -weight => "1");
	$textframe->gridRowconfigure(2,    -weight => "1");
	
	my $lower = $targetwin->{'window'}->Frame;
	$lower->Button(
		-text    => "Get TOC",    
		-command => [\&_get_book_toc, $targetwin, $conwin, $mw]
	)->pack(-side => "left");
	$lower->Button(
		-text    => "Expand",     
		-command => [\&_expand_links, $targetwin]
	)->pack(-side => "left");
	$lower->Button(
		-text    => "Category",   
		-command => [\&_get_cat, $targetwin, $conwin, $mw]
	)->pack(-side => "left");
	$lower->Button(
		-text    => "Links Here", 
		-command => [\&_get_linkshere, $targetwin, $conwin, $mw]
	)->pack(-side => "left");
	$lower->Button(
		-text    => "Prefix",     
		-command => [\&_get_prefixindex, $targetwin, $conwin, $mw]
	)->pack(-side => "left");
	$lower->Button(
		-text => "Unformat",   
		-command => [\&_clean_links, $targetwin]
	)->pack(-side => "left");
	$lower->Button(
		-text => "Clear",      
		-command => [\&_clear_pages_window, $targetwin, $conwin]
	)->pack(-side => "left");
	$textframe->grid(
		-row    => 1, 
		-column => 1, 
		-sticky => "nsew"
	);
	$lower->grid(
		-row    => 2, 
		-column => 1, 
		-sticky => "sw"
	);
	$targetwin->{'window'}->gridRowconfigure(1,    -weight => "1");
	$targetwin->{'window'}->gridColumnconfigure(1, -weight => "1");
	$targetwin->{'window'}->OnDestroy(\sub {$targetwin->{'window'} = undef;});
	$targetwin->{'window'}->title('Page list');
}

sub get_target_pages
{
	my $targetwin = shift;
	return () unless defined $targetwin->{'window'};
	return split("\n", $targetwin->{'pages'}->Contents);
}

sub _clear_pages_window
{
	my ($targetwin, $conwin) = @_;
	return unless defined $targetwin->{'window'};
	$targetwin->{'pages'}->Contents("");
	$targetwin->{'basepage'}->Contents("");
	print_console_message($conwin, "Cleared page list");
}

#remove unnecessary whitespace and line breaks from the page list.
sub _rm_whitespace
{
	my $targetwin = shift;
	my $text = $targetwin->{'pages'}->Contents;
	$text =~ s/\n{2,}/\n/g;		#Kill extra newlines
	$text =~ s/^[ \t]+//g;		#Kill beginning spaces and tabs
	$text =~ s/[ \t]+$//g;		#Kill trailing whitespace
	$targetwin->{'pages'}->Contents($text);
}

sub _get_book_toc
{
	my($targetwin, $conwin, $mw) = @_;
	my $book = $targetwin->{'basepage'}->Contents;
	$book =~ s/\n//g;			#kill all newlines
	my $pg = $mw->get($book);
	my $text = $pg->content();
	my @links = ();
	if(defined($text) && $text ne "")
	{
		while($text =~ m/\[\[([^|\]]+)(|[^\]]+)?\]\]/g)
		{
			push @links, $1;
		}
		$targetwin->{'pages'}->Contents(join("\n", @links));
		_expand_links($targetwin);
	}
	print_console_message($conwin, "Retrieved TOC from '$book', ");
	print_console_message($conwin, "(" . scalar @links . ") pages loaded");
}

sub _clean_links #remove unneeded wikiformatting from links in TOC
{
	my ($targetwin, $conwin) = @_;
	my $text = $targetwin->{'pages'}->Contents;
	my $bookname = $targetwin->{'basepage'}->Contents;
	chomp $bookname;
	my @links = split(/\n/, $text);
	$text = "";
	foreach my $link (@links)
	{
	    $link =~ s/^\W+//g;
		$text = $text . $link . "\n";
	}
	$targetwin->{'pages'}->Contents($text);
}

sub _expand_links #expand the links in the TOC to be fully-qualified
{
	my ($targetwin, $conwin) = @_;
	my $text = $targetwin->{'pages'}->Contents;
	my $basename = $targetwin->{'basepage'}->Contents;
	chomp $basename;
	my $bookname = $basename;
	$bookname =~ s/(.+)(\/.*)/$1/g;
	my @links;
	while($text =~ m/(\.\.\/)?(\/)?([^\n]+)/g)
	{
		my $link;
		if($1)
		{
			$link = $bookname . "/" . $3;
		}
		elsif($2)
		{
			$link = $3;
			$link =~ s/\/$//g;
			$link = $bookname . "/" . $link;
		}
		else
		{
			$link = $3;
		}
		push @links, $link;
	}
	$targetwin->{'pages'}->Contents(join("\n", @links));
}

#Buggy because $mw->readcat() is buggy. Once the library is fixed, this should work well.
sub _get_cat
{
	my ($targetwin, $conwin, $mw) = @_;
	my $cat = $targetwin->{'basepage'}->Contents;
	chomp $cat;
	my ($list, $subcats) = $mw->readcat($cat);
	$targetwin->{'pages'}->Contents(join("\n", @{$list}));
	print_console_message($conwin, "Retrieved list from 'Category:$cat', ");
	print_console_message($conwin, "(" . scalar @{$list} . ") pages loaded");
}

#These just haven't been implemented yet.
sub _get_linkshere
{
	my ($targetwin, $conwin, $mw) = @_;
	print_console_message($conwin, "Links Here not implemented");
	return;
}

sub _get_prefixindex
{
	my ($targetwin, $conwin, $mw) = @_;
	print_console_message($conwin, "Prefixindex not implemented");
	return;
}

sub make_about_window
{
	our %_GLOBALOPTIONS;
	my $mainwin = shift;
	my $about = $mainwin->{'window'}->DialogBox(
		-title   => "About", 
		-buttons => ["Ok"]
	);
	my $stuff = $about->add("ROText",
		-width  => 70, 
		-height => 16, 
		-wrap   => "none",
		-background => $_GLOBALOPTIONS{"ROTEXTBACKGROUND"}, 
		-foreground => $_GLOBALOPTIONS{"ROTEXTFORGROUND"}
	)->pack();
	my $abouttext = << "__END_OF_TEXT__";
Wikperl: Wiki transform editor
Version $VERSION. 
Copyright (C) 2007 Andrew Whitworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

The license is available at http://www.gnu.org/copyleft/gpl.html
__END_OF_TEXT__

	$stuff->Contents($abouttext);
	$about->Show;
}

